@@ -48,8 +48,8 @@ def main_entry(argv):
     public_key_uncompressed_hash160 = b''
     public_key_compressed_hash160 = b''
     public_key_hash160 = b''
-    taproot_internal_key = b''
-    taproot_output_key = b''
+    taproot_tweaked_private_key = b''
+    taproot_tweaked_public_key = b''
     addr_p2pkh_uncompressed = b''
     addr_p2pkh_compressed = b''
     addr_p2pkh = b''  # uncompressed or compressed
@@ -100,18 +100,27 @@ def main_entry(argv):
             sys.stderr.write("derivation path {} is not available for {}\n".format(derivation, chain))
             sys.exit(1)
         private_key = mnemonic_util.mnemonic_to_private_key(mnemonic, derivation_path)
-        private_key_wif = wif_util.encode_wif(private_key, wif_version_bytes)
-        private_key_wif_compressed = wif_util.encode_wif(private_key, wif_version_bytes, compressed_wif=True)
+        if chain == "grs" or chain == "grs_testnet":
+            private_key_wif = wif_util.grs_encode_wif(private_key, wif_version_bytes)
+            private_key_wif_compressed = wif_util.grs_encode_wif(private_key, wif_version_bytes, compressed_wif=True)
+        else:
+            private_key_wif = wif_util.encode_wif(private_key, wif_version_bytes)
+            private_key_wif_compressed = wif_util.encode_wif(private_key, wif_version_bytes, compressed_wif=True)
         public_key_compressed = common_util.prikey_to_pubkey(private_key, compressed=True)
         public_key_compressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_compressed)
         if derivation == "bip44":
             # For legacy address
-            addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+            if chain == "grs" or chain == "grs_testnet":
+                addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+            else:
+                addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
         elif derivation == "bip49":
             # For p2sh-segwit address
             if script_version_bytes:
-                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed,
-                                                                               script_version_bytes)
+                if chain == "grs" or chain == "grs_testnet":
+                    addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
+                else:
+                    addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
         elif derivation == "bip84":
             # For bech32 address
             if human_readable_part:
@@ -119,24 +128,31 @@ def main_entry(argv):
         elif derivation == "bip86":
             # For bech32m address
             if human_readable_part:
+                taproot_tweaked_private_key = p2tr_util.taproot_tweak_seckey(private_key)
                 public_key_x_coordinate = public_key_compressed[1:33]
-                taproot_output_key = p2tr_util.public_key_x_coordinate_to_taproot_output_key(public_key_x_coordinate)
-                addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_output_key)
+                taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
+                addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
         elif derivation.startswith("m/"):  # customized path
             # For legacy address
-            addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+            if chain == "grs" or chain == "grs_testnet":
+                addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+            else:
+                addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
             # For p2sh-segwit address
             if script_version_bytes:
-                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed,
-                                                                               script_version_bytes)
+                if chain == "grs" or chain == "grs_testnet":
+                    addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
+                else:
+                    addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
             # For bech32 address
             if human_readable_part:
                 addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
             # For bech32m address
             if human_readable_part:
+                taproot_tweaked_private_key = p2tr_util.taproot_tweak_seckey(private_key)
                 public_key_x_coordinate = public_key_compressed[1:33]
-                taproot_output_key = p2tr_util.public_key_x_coordinate_to_taproot_output_key(public_key_x_coordinate)
-                addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_output_key)
+                taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
+                addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
     elif (len(inputs) == 66 and inputs.startswith("0x")) or len(inputs) == 64:
         # sys.stderr.write("you input private key\n")
         # private key
@@ -144,21 +160,33 @@ def main_entry(argv):
         # For example: c7ac679b56f50bfd54dd924fe45a8dca7a1c2dced254b03dac22afc03adb9127
         private_key_hex = inputs.lower().replace('0x', '')
         private_key = bytes.fromhex(private_key_hex)
-        private_key_wif = wif_util.encode_wif(private_key, wif_version_bytes)
-        private_key_wif_compressed = wif_util.encode_wif(private_key, wif_version_bytes, compressed_wif=True)
+        if chain == "grs" or chain == "grs_testnet":
+            private_key_wif = wif_util.grs_encode_wif(private_key, wif_version_bytes)
+            private_key_wif_compressed = wif_util.grs_encode_wif(private_key, wif_version_bytes, compressed_wif=True)
+        else:
+            private_key_wif = wif_util.encode_wif(private_key, wif_version_bytes)
+            private_key_wif_compressed = wif_util.encode_wif(private_key, wif_version_bytes, compressed_wif=True)
         public_key_uncompressed = common_util.prikey_to_pubkey(private_key, compressed=False)
         public_key_compressed = common_util.pubkey_uncompressed_to_compressed(public_key_uncompressed)
         public_key_uncompressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_uncompressed)
         public_key_compressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_compressed)
-        addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
-        addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+        if chain == "grs" or chain == "grs_testnet":
+            addr_p2pkh_uncompressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
+            addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+        else:
+            addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
+            addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
         if script_version_bytes:
-            addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
+            if chain == "grs" or chain == "grs_testnet":
+                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
+            else:
+                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
         if human_readable_part:
             addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
+            taproot_tweaked_private_key = p2tr_util.taproot_tweak_seckey(private_key)
             public_key_x_coordinate = public_key_compressed[1:33]
-            taproot_output_key = p2tr_util.public_key_x_coordinate_to_taproot_output_key(public_key_x_coordinate)
-            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_output_key)
+            taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
+            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
     elif (len(inputs) == 130 and inputs.startswith("0x")) or len(inputs) == 128 \
             or (len(inputs) == 132 and inputs.startswith("0x04")) \
             or (len(inputs) == 130 and inputs.startswith("04")):
@@ -173,15 +201,22 @@ def main_entry(argv):
         public_key_compressed = common_util.pubkey_uncompressed_to_compressed(public_key_uncompressed)
         public_key_uncompressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_uncompressed)
         public_key_compressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_compressed)
-        addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
-        addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+        if chain == "grs" or chain == "grs_testnet":
+            addr_p2pkh_uncompressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
+            addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+        else:
+            addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
+            addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
         if script_version_bytes:
-            addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
+            if chain == "grs" or chain == "grs_testnet":
+                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
+            else:
+                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
         if human_readable_part:
             addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
             public_key_x_coordinate = public_key_compressed[1:33]
-            taproot_output_key = p2tr_util.public_key_x_coordinate_to_taproot_output_key(public_key_x_coordinate)
-            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_output_key)
+            taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
+            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
     elif (len(inputs) == 68 and inputs.startswith("0x")) or len(inputs) == 66:
         # sys.stderr.write("you input compressed public key\n")
         # compressed public key
@@ -192,44 +227,72 @@ def main_entry(argv):
         public_key_uncompressed = common_util.pubkey_compressed_to_uncompressed(public_key_compressed)
         public_key_uncompressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_uncompressed)
         public_key_compressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_compressed)
-        addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
-        addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+        if chain == "grs" or chain == "grs_testnet":
+            addr_p2pkh_uncompressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
+            addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+        else:
+            addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
+            addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
         if script_version_bytes:
-            addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
+            if chain == "grs" or chain == "grs_testnet":
+                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
+            else:
+                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
         if human_readable_part:
             addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
             public_key_x_coordinate = public_key_compressed[1:33]
-            taproot_output_key = p2tr_util.public_key_x_coordinate_to_taproot_output_key(public_key_x_coordinate)
-            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_output_key)
+            taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
+            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
     elif (len(inputs) == 42 and inputs.startswith("0x")) or len(inputs) == 40:
         # sys.stderr.write("you input hash160 of public key\n")
         public_key_hash160 = bytes.fromhex(inputs.lower().replace('0x', ''))
-        addr_p2pkh = p2pkh_util.hash160_to_p2pkh_addr(public_key_hash160, pubkey_version_bytes)
+        if chain == "grs" or chain == "grs_testnet":
+            addr_p2pkh = p2pkh_util.grs_hash160_to_p2pkh_addr(public_key_hash160, pubkey_version_bytes)
+        else:
+            addr_p2pkh = p2pkh_util.hash160_to_p2pkh_addr(public_key_hash160, pubkey_version_bytes)
         if script_version_bytes:
-            addr_p2sh_p2wpkh = p2sh_p2wpkh_util.hash160_to_p2sh_p2wpkh_addr(public_key_hash160, script_version_bytes)
+            if chain == "grs" or chain == "grs_testnet":
+                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_hash160_to_p2sh_p2wpkh_addr(public_key_hash160, script_version_bytes)
+            else:
+                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.hash160_to_p2sh_p2wpkh_addr(public_key_hash160, script_version_bytes)
         if human_readable_part:
             addr_p2wpkh = p2wpkh_util.hash160_to_segwit_v0_addr(human_readable_part, public_key_hash160)
-    elif wif_util.is_valid_wif(inputs):
-        if chain == "btc":
+    elif wif_util.is_valid_wif(inputs, chain):
+        if chain == "btc" or chain == "grs":
             assert inputs.startswith('5') or inputs.startswith('K') or inputs.startswith('L')
-        if chain == "btc-test":
+        if chain == "btc-test" or chain == "grs_testnet":
             assert inputs.startswith('9') or inputs.startswith('c')
-        private_key = wif_util.decode_wif(inputs)
-        private_key_wif = wif_util.encode_wif(private_key, wif_version_bytes)
-        private_key_wif_compressed = wif_util.encode_wif(private_key, wif_version_bytes, compressed_wif=True)
+        if chain == "grs" or chain == "grs_testnet":
+            private_key = wif_util.grs_decode_wif(inputs)
+        else:
+            private_key = wif_util.decode_wif(inputs)
+        if chain == "grs" or chain == "grs_testnet":
+            private_key_wif = wif_util.grs_encode_wif(private_key, wif_version_bytes)
+            private_key_wif_compressed = wif_util.grs_encode_wif(private_key, wif_version_bytes, compressed_wif=True)
+        else:
+            private_key_wif = wif_util.encode_wif(private_key, wif_version_bytes)
+            private_key_wif_compressed = wif_util.encode_wif(private_key, wif_version_bytes, compressed_wif=True)
         public_key_uncompressed = common_util.prikey_to_pubkey(private_key, compressed=False)
         public_key_compressed = common_util.pubkey_uncompressed_to_compressed(public_key_uncompressed)
         public_key_uncompressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_uncompressed)
         public_key_compressed_hash160 = p2pkh_util.pubkey_to_hash160(public_key_compressed)
-        addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
-        addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+        if chain == "grs" or chain == "grs_testnet":
+            addr_p2pkh_uncompressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
+            addr_p2pkh_compressed = p2pkh_util.grs_pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
+        else:
+            addr_p2pkh_uncompressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_uncompressed, pubkey_version_bytes)
+            addr_p2pkh_compressed = p2pkh_util.pubkey_to_p2pkh_addr(public_key_compressed, pubkey_version_bytes)
         if script_version_bytes:
-            addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
+            if chain == "grs" or chain == "grs_testnet":
+                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.grs_pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
+            else:
+                addr_p2sh_p2wpkh = p2sh_p2wpkh_util.pubkey_to_p2sh_p2wpkh_addr(public_key_compressed, script_version_bytes)
         if human_readable_part:
             addr_p2wpkh = p2wpkh_util.pubkey_to_segwit_v0_addr(human_readable_part, public_key_compressed)
+            taproot_tweaked_private_key = p2tr_util.taproot_tweak_seckey(private_key)
             public_key_x_coordinate = public_key_compressed[1:33]
-            taproot_output_key = p2tr_util.public_key_x_coordinate_to_taproot_output_key(public_key_x_coordinate)
-            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_output_key)
+            taproot_tweaked_public_key = p2tr_util.public_key_x_coordinate_to_taproot_tweaked_pubkey(public_key_x_coordinate)
+            addr_p2tr = p2wpkh_util.pubkey_to_segwit_v1_addr(human_readable_part, taproot_tweaked_public_key)
     else:
         sys.stderr.write("invalid input: {0}\n".format(inputs))
         sys.exit(1)
@@ -242,6 +305,8 @@ def main_entry(argv):
         print("private key (WIF) = {}".format(private_key_wif.decode('ascii')))
     if private_key_wif_compressed:
         print("private key (WIF compressed) = {}".format(private_key_wif_compressed.decode('ascii')))
+    if taproot_tweaked_private_key:
+        print("taproot tweaked private key = {}".format(taproot_tweaked_private_key.hex()))
     if public_key_uncompressed:
         print("public key (uncompressed) = {}".format(public_key_uncompressed.hex()))
     if public_key_compressed:
@@ -252,8 +317,8 @@ def main_entry(argv):
         print("hash160 of compressed public key = {}".format(public_key_compressed_hash160.hex()))
     if public_key_hash160:
         print("hash160 of public key = {}".format(public_key_hash160.hex()))
-    if taproot_output_key:
-        print("taproot output key = {}".format(taproot_output_key.hex()))
+    if taproot_tweaked_public_key:
+        print("taproot tweaked public key (taproot output key) = {}".format(taproot_tweaked_public_key.hex()))
     if addr_p2pkh_uncompressed:
         print("legacy address (p2pkh uncompressed) = {}".format(addr_p2pkh_uncompressed.decode('ascii')))
         if chain == "bch":
